From: Michal Pecio <michal.pecio@gmail.com>
To: Mathias Nyman <mathias.nyman@intel.com>,
Greg Kroah-Hartman <gregkh@linuxfoundation.org>
Cc: Bart Nagel <bart@tremby.net>,
linux-usb@vger.kernel.org, linux-kernel@vger.kernel.org
Subject: [PATCH 5/5] usb: xhci: Rework and improve the TD matching and skipping logic
Date: Tue, 4 Aug 2026 12:05:37 +0200 [thread overview]
Message-ID: <20260804120537.5c30554e.michal.pecio@gmail.com> (raw)
In-Reply-To: <20260804120110.01bda0e2.michal.pecio@gmail.com>
Matching events with TDs and giving back missed TDs is carried out
by a complicated loop. Replace it with a simpler linear logic:
0. Having verified that 'td_list' isn't empty,
1. Scan it to find the matching TD and count missed TDs,
2. Perform necessary adjustments for corner cases,
3. Give back missed TDs, if applicable, using a short and tidy loop,
4. Check if the event refers to the expected TD and proceed as usual.
Besides cleaning up the code, this provides a few improvements:
- when the skip flag is set, no TD is given back unless we found a match
or otherwise know how many TDs should be given back
- when the skip flag is clear, we know if the event refers to a "future"
TD so we can log this in the Scary Error Message to aid debugging.
While altering the error message, drop a pointless goto.
Signed-off-by: Michal Pecio <michal.pecio@gmail.com>
---
drivers/usb/host/xhci-ring.c | 137 ++++++++++++++++-------------------
1 file changed, 61 insertions(+), 76 deletions(-)
diff --git a/drivers/usb/host/xhci-ring.c b/drivers/usb/host/xhci-ring.c
index 1f0cb6a701c5..d94146ceb178 100644
--- a/drivers/usb/host/xhci-ring.c
+++ b/drivers/usb/host/xhci-ring.c
@@ -125,11 +125,16 @@ static bool link_trb_toggles_cycle(union xhci_trb *trb)
return le32_to_cpu(trb->link.control) & LINK_TOGGLE;
}
-static bool last_td_in_urb(struct xhci_td *td)
+static int num_tds_not_done(struct urb *urb)
{
- struct urb_priv *urb_priv = td->urb->hcpriv;
+ struct urb_priv *urb_priv = urb->hcpriv;
- return urb_priv->num_tds_done == urb_priv->num_tds;
+ return urb_priv->num_tds - urb_priv->num_tds_done;
+}
+
+static bool last_td_in_urb(struct xhci_td *td)
+{
+ return !num_tds_not_done(td->urb);
}
static bool unhandled_event_trb(struct xhci_ring *ring)
@@ -2605,14 +2610,19 @@ static bool xhci_spurious_success_tx_event(struct xhci_hcd *xhci,
}
}
-static struct xhci_td *find_td_by_dma(struct xhci_ring *ep_ring, dma_addr_t dma)
+static struct xhci_td *find_td_by_dma(struct xhci_ring *ep_ring, int *missed_tds, dma_addr_t dma)
{
struct xhci_td *td;
- if (dma)
+ if (dma) {
list_for_each_entry(td, &ep_ring->td_list, td_list)
if (trb_in_td(td, dma))
return td;
+ else
+ (*missed_tds)++;
+ }
+
+ *missed_tds = 0;
return NULL;
}
@@ -2629,8 +2639,8 @@ static int handle_tx_event(struct xhci_hcd *xhci,
struct xhci_ring *ep_ring;
unsigned int slot_id;
int ep_index;
- struct xhci_td *td = NULL;
- struct urb *missed_urb = NULL;
+ struct xhci_td *td;
+ int missed_tds = 0;
dma_addr_t ep_trb_dma;
union xhci_trb *ep_trb;
int status = -EINPROGRESS;
@@ -2813,13 +2823,6 @@ static int handle_tx_event(struct xhci_hcd *xhci,
xhci_dequeue_td(xhci, td, ep_ring, td->status);
}
- /*
- * We don't know how many TDs were missed when ep_trb_dma is zero (as permitted by
- * xHCI 1.0) or bogus. Bail out leaving ep->skip set, next event will sort it out.
- */
- if (trb_comp_code == COMP_MISSED_SERVICE_ERROR && !find_td_by_dma(ep_ring, ep_trb_dma))
- return 0;
-
if (list_empty(&ep_ring->td_list)) {
/*
* Don't print wanings if ring is empty due to a stopped endpoint generating an
@@ -2839,63 +2842,47 @@ static int handle_tx_event(struct xhci_hcd *xhci,
goto check_endpoint_halted;
}
- do {
- td = list_first_entry(&ep_ring->td_list, struct xhci_td,
- td_list);
-
- if (ep->skip) {
-
- if (!trb_in_td(td, ep_trb_dma)) {
- /* this event is unlikely to match any TD, don't skip them all */
- if (trb_comp_code == COMP_STOPPED_LENGTH_INVALID)
- return 0;
-
- /*
- * If skip flag is still set at xrun, we are on xHCI 1.0 and our TRB
- * pointer is zero again. All missed TDs can be given back, but we
- * don't know which were missed and which were queued after the xrun
- * occurred. We can safely give back the first pending URB.
- */
- if (ring_xrun_event) {
- if (!missed_urb)
- missed_urb = td->urb;
-
- if (td->urb != missed_urb) {
- xhci_dbg(xhci, "Skipped one URB for slot %u ep %u",
- slot_id, ep_index);
- return 0;
- }
- }
-
- /*
- * TD was missed, skip it. Core already initialized frame->status
- * to -EXDEV and frame->actual_length to 0, nothing more to do.
- */
- xhci_dequeue_td(xhci, td, ep_ring, 0);
+ td = find_td_by_dma(ep_ring, &missed_tds, ep_trb_dma);
- if (!list_empty(&ep_ring->td_list))
- continue;
+ if (ep->skip) {
+ if (!td) {
+ /*
+ * xHCI 1.0 allowed MSE events to have zero TRB pointers. Some old chips
+ * also generate bogus non-zero pointers. We know, don't bother warning.
+ * Missed TDs will be given back by the next event with a valid pointer.
+ */
+ if (trb_comp_code == COMP_MISSED_SERVICE_ERROR &&
+ xhci->hci_version <= 0x100)
+ return 0;
+ /*
+ * If skip flag is still set at xrun, we are on xHCI 1.0 and our TRB pointer
+ * is zero again. All missed TDs can be given back, but we don't know which
+ * were missed and which were queued after the xrun occurred. We can safely
+ * give back the first pending URB to let the class driver know.
+ */
+ if (ring_xrun_event)
+ missed_tds = num_tds_not_done(list_first_entry(&ep_ring->td_list,
+ struct xhci_td, td_list)->urb);
+ /* In other cases missed_tds is zero */
+ }
- xhci_dbg(xhci, "All TDs skipped for slot %u ep %u. Clear skip flag.\n",
- slot_id, ep_index);
- ep->skip = false;
- td = NULL;
- goto check_endpoint_halted;
- }
+ /*
+ * Give back missed TDs. Core already initialized their frame->status to -EXDEV
+ * and frame->actual_length to 0, nothing more to do.
+ */
+ for (int i = 0; i < missed_tds; i++)
+ xhci_dequeue_td(xhci,
+ list_first_entry(&ep_ring->td_list, struct xhci_td, td_list),
+ ep_ring, 0);
- xhci_dbg(xhci,
- "Found td. Clear skip flag for slot %u ep %u.\n",
- slot_id, ep_index);
+ /* the list may become empty on ring_xrun_event */
+ if (td || list_empty(&ep_ring->td_list))
ep->skip = false;
- }
- /*
- * If ep->skip is set, it means there are missed tds on the
- * endpoint ring need to take care of.
- * Process them as short transfer until reach the td pointed by
- * the event.
- */
- } while (ep->skip);
+ xhci_dbg(xhci, "Skipped %d TDs on slot %u ep %u comp_code %u, TD found %d, skip flag %d\n",
+ missed_tds, slot_id, ep_index, trb_comp_code, !!td, ep->skip);
+ missed_tds = 0;
+ }
ep_ring->old_trb_comp_code = trb_comp_code;
@@ -2907,7 +2894,7 @@ static int handle_tx_event(struct xhci_hcd *xhci,
return 0;
/* Handle events not referencing the current TD */
- if (!trb_in_td(td, ep_trb_dma)) {
+ if (!td || missed_tds) {
/*
* Skip the Force Stopped Event. The 'ep_trb' of FSE is not in the current
* TD pointed by 'ep_ring->dequeue' because that the hardware dequeue
@@ -2930,7 +2917,13 @@ static int handle_tx_event(struct xhci_hcd *xhci,
}
/* HC is busted, give up! */
- goto debug_finding_td;
+ td = list_first_entry(&ep_ring->td_list, struct xhci_td, td_list);
+ xhci_err(xhci, "Event dma %pad for ep %d comp_code %u not part of TD at %016llx - %016llx, missed %d\n",
+ &ep_trb_dma, ep_index, trb_comp_code,
+ (u64)xhci_trb_virt_to_dma(td->start_seg, td->start_trb),
+ (u64)xhci_trb_virt_to_dma(td->end_seg, td->end_trb),
+ missed_tds);
+ return -ESHUTDOWN;
}
trace_xhci_handle_transfer(ep_ring, (struct xhci_generic_trb *) ep_trb, ep_trb_dma);
@@ -2962,14 +2955,6 @@ static int handle_tx_event(struct xhci_hcd *xhci,
return 0;
-debug_finding_td:
- xhci_err(xhci, "Event dma %pad for ep %d status %d not part of TD at %016llx - %016llx\n",
- &ep_trb_dma, ep_index, trb_comp_code,
- (unsigned long long)xhci_trb_virt_to_dma(td->start_seg, td->start_trb),
- (unsigned long long)xhci_trb_virt_to_dma(td->end_seg, td->end_trb));
-
- return -ESHUTDOWN;
-
err_out:
xhci_err(xhci, "@%016llx %08x %08x %08x %08x\n",
(unsigned long long) xhci_trb_virt_to_dma(
--
2.48.1
next prev parent reply other threads:[~2026-08-04 10:05 UTC|newest]
Thread overview: 15+ messages / expand[flat|nested] mbox.gz Atom feed top
2026-08-04 10:01 [PATCH 0/5] xhci: Sort out the TD skipping business Michal Pecio
2026-08-04 10:02 ` [PATCH 1/5] usb: xhci: Handle bogus TRB pointers in Missed Service Error events Michal Pecio
2026-08-05 17:13 ` Mathias Nyman
2026-08-04 10:03 ` [PATCH 2/5] usb: xhci: Guarantee URB giveback on Ring Underrun/Overrun Michal Pecio
2026-08-04 10:03 ` [PATCH 3/5] usb: xhci: Don't set the skip flag on non-isoc endpoints Michal Pecio
2026-08-04 10:04 ` [PATCH 4/5] usb: xhci: Shorten the TD skipping loop Michal Pecio
2026-08-05 17:14 ` Mathias Nyman
2026-08-04 10:05 ` Michal Pecio [this message]
2026-08-05 17:39 ` [PATCH 5/5] usb: xhci: Rework and improve the TD matching and skipping logic Mathias Nyman
2026-08-05 19:30 ` Michal Pecio
2026-08-06 11:00 ` Michal Pecio
2026-08-06 22:16 ` Mathias Nyman
2026-08-05 19:06 ` [PATCH 0/5] xhci: Sort out the TD skipping business Bart Nagel
2026-08-05 20:21 ` Michal Pecio
2026-08-06 18:28 ` Bart Nagel
Reply instructions:
You may reply publicly to this message via plain-text email
using any one of the following methods:
* Save the following mbox file, import it into your mail client,
and reply-to-all from there: mbox
Avoid top-posting and favor interleaved quoting:
https://en.wikipedia.org/wiki/Posting_style#Interleaved_style
* Reply using the --to, --cc, and --in-reply-to
switches of git-send-email(1):
git send-email \
--in-reply-to=20260804120537.5c30554e.michal.pecio@gmail.com \
--to=michal.pecio@gmail.com \
--cc=bart@tremby.net \
--cc=gregkh@linuxfoundation.org \
--cc=linux-kernel@vger.kernel.org \
--cc=linux-usb@vger.kernel.org \
--cc=mathias.nyman@intel.com \
/path/to/YOUR_REPLY
https://kernel.org/pub/software/scm/git/docs/git-send-email.html
* If your mail client supports setting the In-Reply-To header
via mailto: links, try the mailto: link
Be sure your reply has a Subject: header at the top and a blank line
before the message body.
This is a public inbox, see mirroring instructions
for how to clone and mirror all data and code used for this inbox